import abc
from typing import Union
from cdocs.contextual_docs import DocPath, FilePath

class Filer(metaclass=abc.ABCMeta):
    """
    Filer recognizes file types.
    """

    @abc.abstractmethod
    def get_filetype( self, path:Union[DocPath,FilePath]) -> str:
        pass

    @abc.abstractmethod
    def is_probably_not_binary(self, path:Union[DocPath,FilePath] ) -> bool:
        pass




import abc
from typing import Optional, List
from cdocs.contextual_docs import Doc, DocPath, JsonDict


class MultiContextDocs(metaclass=abc.ABCMeta):
    """
    MultiContextDocs retrieves docs and labels from one
    or more Cdocs named in a list.
    """

    @abc.abstractmethod
    def get_root_names_accepting_path(self, path:DocPath) -> List[str]:
        """
        returns root names that work with docpath in the get_docs method
        based on the doc's extension.
            /x/y/z would work with any root taking "cdocs".
            /x/y/z.html would work with any root taking "html".
        note that any docpath with a dotted extension is unacceptable for
        get_docs in a "cdocs" root. docpath with dotted extensions are
        fine for get_docs on roots that are declared to handle other formats.
        ".config" and dotted extension compose docs have methods that
        permit them in "cdocs" roots.
        """
        pass

    @abc.abstractmethod
    def get_concat_doc_from_roots(self, rootnames:List[str], path:DocPath) -> Optional[Doc]:
        pass

    @abc.abstractmethod
    def get_compose_doc_from_roots(self, rootnames:List[str], path:DocPath) -> Optional[Doc]:
        pass

    @abc.abstractmethod
    def get_doc_from_roots(self, rootnames:List[str], path:DocPath, \
                           notfound:Optional[bool]=True, \
                           splitplus:Optional[bool]=True) -> Optional[Doc]:
        pass

    @abc.abstractmethod
    def get_labels_from_roots(self, rootnames:List[str], path:DocPath, recurse:Optional[bool]=True) ->  Optional[JsonDict]:
        pass

    @abc.abstractmethod
    def get_tokens_from_roots(self, rootnames:List[str], path:DocPath, recurse:Optional[bool]=True) ->  Optional[JsonDict]:
        pass

    @abc.abstractmethod
    def list_docs_from_roots(self, rootnames:List[str], path:DocPath) ->  Optional[JsonDict]:
        pass



import abc
from cdocs.contextual_docs import Doc, DocPath
from typing import List


class Concatter(metaclass=abc.ABCMeta):
    """
    Concatter knows how to string content together into a single string.
    """

    @abc.abstractmethod
    def concat(self, paths:List[DocPath]) -> Doc:
        path




from typing import Optional, List, Dict
from datetime import datetime
from jinja2 import Template
import logging
import os
from cdocs.contextual_docs import Doc, DocPath, FilePath, JsonDict, ContextualDocs
from cdocs.config import Config, ConfigException
from cdocs.pather import Pather
from cdocs.reader import Reader
from cdocs.lister import Lister
from cdocs.finder import Finder
from cdocs.filer import Filer
from cdocs.transformer import Transformer
from cdocs.concatter import Concatter
from cdocs.simple_transformer import SimpleTransformer
from cdocs.simple_concatter import SimpleConcatter
from cdocs.simple_config import SimpleConfig
from cdocs.simple_reader import SimpleReader
from cdocs.simple_lister import SimpleLister
from cdocs.simple_pather import SimplePather
from cdocs.simple_finder import SimpleFinder
from cdocs.simple_filer import SimpleFiler
from cdocs.physical import Physical
from cdocs.multi_context_docs import MultiContextDocs
from cdocs.context_metadata import ContextMetadata
from cdocs.changer import Changer

class DocNotFoundException(Exception):
    pass

class BadDocPath(Exception):
    pass

class BadFilePath(Exception):
    pass

class ComposeDocException(Exception):
    pass


class Cdocs(ContextualDocs, Physical, Changer):

    def __init__(self, docspath:str, config:Optional[Config]=None, context:Optional[MultiContextDocs]=None):
        super().__init__()
        cfg = SimpleConfig(None) if config is None else config
        self._config = cfg
        self._context:MultiContextDocs = context
        self._docs_path:FilePath = docspath
        self._rootname = cfg.get_matching_key_for_value("docs", docspath)
        if self._rootname is None:
            raise ConfigException(f"Cdocs.__init__: no rootname for {docspath}")
        self._set_ext()
        self._tokens_filename:str  = cfg.get("filenames", "tokens", "tokens.json")
        self._labels_filename:str  = cfg.get("filenames", "labels", "labels.json")
        self._hashmark:str  = cfg.get("filenames", "hashmark", "#")
        self._plus:str  = cfg.get("filenames", "plus", "+")
        self._accepts = None

        self._track_last_change = False
        self._last_change = None
        self._last_change = self.get_last_change()
        #
        # these helpers can be swapped in and out as needed
        #
        self._concatter = SimpleConcatter(self)
        self._lister = SimpleLister(self)
        self._filer = SimpleFiler(self)
        self._transformer = SimpleTransformer(self)
        self._reader = SimpleReader(self) if cfg.reader is None else cfg.reader
        self._finder = SimpleFinder(self) if cfg.finder is None else cfg.finder

        if cfg.pather is None:
            logging.info("Cdocs.__init__: config's pather is None: {cfg}")
            metadata = ContextMetadata(cfg)
            #metadata.config = cfg
            self._pather = SimplePather(metadata, self)
            pass
        else:
            self._pather = cfg.pather
        logging.info(f"Cdocs.__init__: completed init. Cdocs is:\n{str(self)}")

    def __str__(self):
        return  f"Cdocs: type(self): "+\
                f"path: {self._docs_path}, "+\
                f"accepts: {self.accepts}, "+\
                f"exts: {self._exts}, "+\
                f"tokens: {self._tokens_filename}, "+\
                f"labels: {self._labels_filename}, "+\
                f"hash: {self._hashmark}, "+\
                f"plus: {self._plus}, "+\
                f"config: {self._config}, "+\
                f"context: {self._context}"+\
                f"concatter: {self._concatter}, "+\
                f"lister: {self._lister}, "+\
                f"filer: {self._filer}, "+\
                f"transformer: {self._transformer}, "+\
                f"reader: {self._reader}, "+\
                f"finder: {self._finder}, "+\
                f"pather: {self._pather}"

    def _set_ext(self) -> None:
        ext = self.config.get("defaults", "ext", "xml")
        logging.info(f"cdocs._set_ext 1: {ext}")
        ext = self.config.get("formats", self.rootname, ext)
        logging.info(f"cdocs._set_ext 2: {ext}")
        if ext.find(",") > -1:
            self._exts = ext.split(",")
        else:
            self._exts = [ext]

    def get_doc_root(self) -> FilePath:
        return FilePath(self._docs_path)

    @property
    def concatter(self) -> Concatter:
        return self._concatter

    @concatter.setter
    def concatter(self, c:Concatter) -> None:
        self._concatter = c

    @property
    def reader(self) -> Reader:
        return self._reader

    @reader.setter
    def reader(self, val:Reader) -> None:
        self._reader = val

    @property
    def finder(self) -> Finder:
        return self._finder

    @finder.setter
    def finder(self, val:Finder) -> None:
        self._finder = val

    @property
    def pather(self) -> Pather:
        return self._pather

    @pather.setter
    def pather(self, val:Pather) -> None:
        self._pather = val

    @property
    def lister(self) -> Lister:
        return self._lister

    @lister.setter
    def lister(self, val:Lister) -> None:
        self._lister = val

    @property
    def context(self) -> MultiContextDocs:
        return self._context

    @context.setter
    def context(self, val:MultiContextDocs) -> None:
        self._context = val

    @property
    def transformer(self) -> Transformer:
        return self._transformer

    @transformer.setter
    def transformer(self, val:Transformer) -> None:
        self._transformer = val

    @property
    def filer(self):
        return self._filer

    @filer.setter
    def filer(self, val:Filer) -> None:
        self._filer = val

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, val:Config) -> None:
        self._config = val

    @property
    def rootname(self):
        return self._rootname

    @property
    def exts(self):
        return self._exts

    @property
    def accepts(self):
        if self._accepts is None:
            logging.info(f"Cdocs.accepts: rootname: {self.rootname}")
            a = self.config.get("accepts", self.rootname)
            if a is None:
                self._accepts = ["cdocs"]
            else:
                self._accepts = a.split(",")
            logging.info(f"Cdocs.accepts: accepts: {self._accepts}")
        return self._accepts

# ===================
# abc methods
# ===================

    def get_last_change(self) -> datetime:
        if self.track_last_change:
            if self._last_change is None:
                lcf = self._get_last_change_file_path()
                if os.path.exists(lcf):
                    last = None
                    with( open(lcf,"r") ) as lc:
                        last = lc.read()
                    self._last_change = datetime.fromtimestamp(float(last))
                else:
                    self.set_last_change()
            return self._last_change

    def _get_last_change_file_path(self):
        if self.track_last_change:
            path = f"{self.get_doc_root()}/.last_change"
            if not os.path.exists(path) :
                try:
                    os.makedirs(self.get_doc_root())
                except Exception as e:
                    logging.error("Cdocs._get_last_change_file_path: create root failed: {self.get_doc_root()}: {e}")
            return path

    def set_last_change(self) -> None:
        if self.track_last_change:
            self._last_change = datetime.now()
            self._write_last_change()

    def _write_last_change(self):
        if self.track_last_change:
            # write the datetime
            seconds = self._last_change.timestamp()
            # write
            lcf = self._get_last_change_file_path()
            with( open(lcf, "w") ) as lc:
                try:
                    lc.write(str(seconds))
                except Exception as e:
                    logging.warning(f"cdocs._write_last_change: cannot write last change. refusing to crash.  {e}")

    @property
    def track_last_change(self) -> bool:
        return self._track_last_change

    @track_last_change.setter
    def track_last_change(self, do:bool) -> None:
        self._track_last_change = do

    def reset_last_change(self, dt:datetime, root:Optional[str]=None) -> None:
        if self.track_last_change:
            if root is not None and self.root_name != root:
                logging.info("Cdocs.reset_last_change: not for me: {root}")
                return
            self._last_change = dt
            self._write_last_change()

    def get_tokens(self, path:DocPath, recurse:Optional[bool]=True, addlabels:Optional[bool]=True) -> JsonDict:
        tokens = self._get_dict(path, self._tokens_filename, recurse)
        if addlabels:
            tokens = self._add_labels_to_tokens(path, tokens, recurse)
        return tokens


    def list_docs(self, path:DocPath) -> List[Doc]:
        return self.lister.list_docs(path)

    def get_labels(self, path:DocPath, recurse:Optional[bool]=True) -> JsonDict:
        labels = self._get_dict(path, self._labels_filename, recurse)
        return self._transform_labels(path, labels)

    def get_compose_doc(self, path:DocPath) -> Doc:
        if path is None :
            raise DocNotFoundException("path can not be None")
        filepath:FilePath = self._pather.get_full_file_path(path)
        try:
            print(f"Cdocs.get_compose_doc: fp: {filepath}")
            content = self._read_doc(filepath)
            tokens:dict = self.get_tokens(path[0:path.rindex('/')])
            content = self.transformer.transform(content, path, tokens, True)
            return Doc(content)
        except Exception as e:
            logging.error(f"Cdocs.get_compose_doc: cannot compose {path}: {e}")
            raise ComposeDocException(f'{path} failed to compose')

    def get_concat_doc(self, path:DocPath) -> Doc:
        if path is None :
            raise DocNotFoundException("path can not be None")
        if path.find('.concat') == -1:
            raise BadDocPath("path must have a .concat file extension")
        paths = self._get_concat_paths(path)
        if paths is None:
            raise DocNotFoundException(f'No concat instruction file at {path}')
        content = self.concatter.concat(paths)
        return Doc(content)

    def get_doc(self, path:DocPath, notfound:Optional[bool]=True) -> Doc:
        return self._get_doc(path, notfound)

    def _get_doc(self, path:DocPath, notfound) -> Optional[Doc]:
        logging.info(f"Cdocs._get_doc: looking for path: {path} in root: {self.rootname}. notfound: {notfound}")
        if path is None :
            raise DocNotFoundException("path can not be None")
        if path.find('.') > -1:
            if self.filer.get_filetype(path) == 'cdocs':
                raise BadDocPath("dots are not allowed in cdoc paths")
        if notfound is None:
            logging.info("Cdocs._get_doc: notfound is None. you should fix this unless you want None returns.")
            notfound = False
        logging.info(f"Cdocs._get_doc: path: {path}")
        pluspaths = self._get_plus_paths(path)
        logging.info(f"Cdocs._get_doc: pluspaths to concationate: {pluspaths}")
        root = self.get_doc_root()
        logging.info(f"Cdocs._get_doc: root {root}")
        doc = self._get_doc_for_root(path, pluspaths, root)
        if doc is None and notfound:
           doc = self.get_404()
        return doc

# ===================
# internal methods
# ===================

    def get_404(self) -> Optional[Doc]:
        config = self.config
        _404 = config.get("notfound", self._rootname, None)
        logging.info(f"Cdocs.get_404: notfound in {self._rootname} is {_404}")
        if _404 is None:
            return None
        doc = self.get_doc(_404, False)
        if doc is None:
            logging.error(f"Cdocs.get_404: {self._rootname}'s notfound: {_404} is None. you shouldl fix this.")
        logging.info(f"Cdocs.get_404: doc: {doc}")
        return doc

    def _get_doc_for_root(self, path:DocPath, pluspaths:List[DocPath], root:FilePath) -> Doc:
        logging.info(f"Cdocs._get_doc_for_root: path: {path}. plus paths: {pluspaths}. root: {root}")
        if len(pluspaths) > 0:
            logging.info(f"Cdocs._get_doc_for_root: stripping down base path from plus path(s)")
            plus = path.find(self._plus)
            path = path[0:plus]
            logging.info(f"Cdocs._get_doc_for_root: base path is now {path}")
        logging.info(f"Cdocs._get_doc_for_root: checking pather: {self._pather} for path: {path}")
        filepath = self._pather.get_full_file_path_for_root(path, root)
        logging.info(f"Cdocs._get_doc_for_root: filepath from pather: {filepath}")
        content = self._read_doc(filepath)
        logging.info(f"Cdocs._get_doc_for_root: content from {filepath} is {len(content) if content is not None else 0} chars. transforming with: {self.transformer}.")
        content = self.transformer.transform(content, path, None, True)
        if len(pluspaths) > 0:
            content = self.concatter.join(content, self.concatter.concat(pluspaths))
        return Doc(content)

    def _transform_labels(self, path:DocPath, labels:JsonDict) -> JsonDict:
        tokens:dict = self.get_tokens(path, addlabels=False)
        ls = { k:self.transformer.transform(v, path, tokens, False) for k,v in labels.items() }
        return JsonDict(ls)

    def _add_labels_to_tokens(self, path:DocPath, tokens:JsonDict, recurse:Optional[bool]=True) -> JsonDict:
        apath = path
        if path.find(self._hashmark) > -1:
            apath = apath[0:apath.find(self._hashmark)]
        if apath.find(self._plus) > -1:
            apath = apath[0:apath.find(self._plus)]
        labels = self.get_labels(apath, recurse)
        ltokens = { "label__"+k:v for k,v in labels.items()}
        tokens  = {**ltokens, **tokens}
        return JsonDict(tokens)

    def _get_plus_paths( self, path:DocPath) -> List[DocPath]:
        lines = path.split(self._plus)
        if len(lines) == 0:
            return lines
        first = lines[0]
        lines = lines[1:]
        mark = first.find(self._hashmark)
        if mark > -1:
            first = first[0:mark+1]
        else:
            first += self._hashmark
        lines = [ DocPath(first+line) for line in lines]
        return lines

    def _get_concat_paths(self, path:DocPath) -> Optional[List[DocPath]]:
        filepath = self._pather.get_full_file_path(path)
        try:
            content = self._read_doc(filepath)
            lines = [DocPath(line) for line in content.split('\n')]
            return lines
        except DocNotFoundException:
            logging.warn(f"Cdocs._get_concat_paths: No such doc {path}. returning None.")
            return None

    def _read_doc(self, path:FilePath) -> str:
        content = None
        available = self._reader.is_available(path)
        logging.info(f"Cdocs._read_doc: {path} is available: {available}")
        if available:
            content = self.reader.read(path)
            if self.filer.is_probably_not_binary(path):
                content = content.decode('utf-8')
            if content is None:
                logging.warning(f"Cdocs._read_doc: cannot read {path}. returning None.")
        else:
            logging.debug(f"Cdocs._read_doc: No such doc {path}. returning None.")
        logging.info(f"Cdocs._read_doc: returning: {content}")
        return content

    def _get_dict(self, path:str, filename:str, recurse:Optional[bool]=True) -> JsonDict:
        path = path.strip('/\\')
        return JsonDict(self.finder.find_tokens(path, filename, recurse))






from typing import List
from cdocs.contextual_docs import Doc, DocPath
from cdocs.lister import Lister
import os
import logging

class SimpleLister(Lister):

    def __init__(self, cdocs):
        self._cdocs = cdocs

    def list_docs(self, path:DocPath) -> List[Doc]:
        logging.info(f"SimpleLister.list_docs: path: {path}")
        #
        # a docpath points to a directory which may have a sibling file named
        # the same as the directory, but with a file extension.
        #
        # listing the docpath means listing the parent directory's files
        # so docpath:
        #    /x/y/z
        # would list:
        #    /x/y
        # and might return:
        #    /x/y/z.xml
        #    /x/y/fish.xml
        #    /x/y/bats.xml
        # it would not return anything anything like:
        #    /x/y/z/z.xml
        #    /x/y/z/fish.xml
        #    /x/y/z/bats.xml
        #
        # why does it work that way? because /x/y/z is an identifier
        # in concept-space, not a filesystem path. as a concept, it includes
        # materials that explain the space. the materials may require
        # multiple files that are all grouped by the concept /x/y/z. we
        # address the fish.xml and bats.xml files as:
        #    /x/y/z#fish.xml
        #    /x/y/z#bats.xml
        #
        # one way to think about it is that z.xml is the important thing and
        # 'z' the directory is just an indication that the concept described
        # by z.xml contains more things.
        #
        # if /x/y/z/a.xml doesn't have a sibling 'a' directory, is /x/y/z/a
        # still good docpath? yes. that just means that the concept
        # /x/y/z/a doesn't contain any further concepts.
        #
        # all this nets out that z.xml, fish.xml and bats.xml just
        # describe /x/y/z.
        #
        # is that a good way to do it? sure. it works, it makes sense
        # logically and you can get used to it pretty easily.
        #
        apath = path
        if path[0:1] == '/' and path.count('/') >=2 or path[0:1] != '/' and path.count('/') >=1:
            apath = path[0:apath.rindex('/')]
        logging.info(f"SimpleLister.list_docs: apath: {apath}")
        root_path = self._cdocs.get_doc_root()
        logging.info(f"SimpleLister.list_docs: root: {root_path}")
        if apath[0:1] == '/':
            apath = apath[1:]
        the_path = os.path.join(root_path, apath)
        logging.info(f"SimpleLister.list_docs: the_path: {the_path}")
        if os.path.exists(the_path):
            logging.info(f"SimpleLister.list_docs: path exists")
            if not os.path.isdir(the_path):
                logging.info(f"SimpleLister.list_docs: the_path: {the_path} is a directory")
                name = the_path[the_path.rindex("/")+1:]
                logging.info(f"SimpleLister.list_docs: returning name: {name}")
                return [name]
            else:
                files = os.listdir(the_path)
                logging.info(f"SimpleLister.list_docs: files: {files}")
                files = [f for f in files if f[0:1] != '.' and os.path.isfile(os.path.join(the_path, f))]
                logging.info(f"SimpleLister.list_docs: returning: {files}")
                return files
        else:
            logging.info(f"SimpleLister.list_docs: directory at {the_path} doesn't exist. no files found. returning [].")
            return []




from cdocs.simple_config import SimpleConfig
import unittest

class ConfigTests(unittest.TestCase):

    noise = False
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def off(self) -> bool:
        return False

    def test_get_config_items(self):
        self._print(f"ConfigTests.test_get_config_items")
        if self.off(): return
        cfg = SimpleConfig()
        items = cfg.get_items("docs", ["public"])
        self._print(f"test_get_config_items: items: {items}")
        self.assertEqual( len(items), 4, msg=f"must be 3 items for docs, not {items}" )
        items = cfg.get_items("fish")
        self.assertEqual( len(items), 0, msg=f"must be 0 items for fish, not {items}" )

    def test_get_with_default(self):
        self._print(f"ConfigTests.test_get_with_default")
        if self.off(): return
        cfg = SimpleConfig()
        oh = cfg.get("fish", "bats", "yeah!")
        self._print(f"test_get_with_default: {oh}")
        self.assertEqual( oh, "yeah!", msg="must equal the default value" )

    def test_get_matching_key_for_value(self):
        self._print(f"ConfigTests.test_get_matching_key_for_value")
        if self.off(): return
        cfg = SimpleConfig()
        name = cfg.get_matching_key_for_value("docs", "/Users/davidkershaw/dev/cdocs/docs/internal")
        self._print(f"ConfigTests.test_get_matching_key_for_value: name1: {name}")
        self.assertEqual( name, "internal", msg="must equal the 'internal' root" )
        name = cfg.get_matching_key_for_value("docs", "fish")
        self._print(f"ConfigTests.test_get_matching_key_for_value: name2: {name}")
        self.assertNotEqual( name, "internal", msg="must not equal the 'internal' root" )




from typing import Optional, List, Dict
import logging
from cdocs.config import Config
from cdocs.simple_config import SimpleConfig
from cdocs.contextual_docs import FilePath


class ContextMetadata(object):

    def __init__(self, config:Optional[Config]=None):
        self._config = SimpleConfig() if config is None else config
        self._roots:List[str] = [ _[1] for _ in self.config.get_items("docs")]
        self._keyed_roots = { _[0]:_[1] for _ in self.config.get_items("docs")}
        self._root_names = [ _[0] for _ in self.config.get_items("docs")]
        self._accepts = {_[0]:[s for s in _[1].split(",")] for _ in self.config.get_items("accepts")}
        self._accepted_by = dict()
        self._load_accepted_by()
        """for k in self._accepts:
            v = self._accepts[k]
            for av in v:
                acceptors = self._accepted_by.get(av)
                if acceptors is None:
                    acceptors = []
                acceptors.append(k)
                self._accepted_by[av] = acceptors
        """
        self._formats = {_[0]:[s for s in _[1].split(",")] for _ in self.config.get_items("formats")}
        self._uses_format = dict()
        for k in self._formats:
            v = self._formats[k]
            for av in v:
                fs = self._uses_format.get(av)
                if fs is None:
                    fs = []
                fs.append(k)
                self._uses_format[av] = fs

    def _load_accepted_by(self):
        logging.info(f"ContextMetadata._load_accepted_by: starting with: {self._accepts}")
        for k,v in self._accepts.items():
            #v = self._accepts[k]
            for av in v:
                acceptors = self._accepted_by.get(av)
                if acceptors is None:
                    acceptors = []
                acceptors.append(k)
                self._accepted_by[av] = acceptors


    @property
    def accepted_by(self) -> Dict[str,List[str]]:
        return self._accepted_by

    @property
    def accepts(self) -> Dict[str,List[str]]:
        return self._accepts

    @property
    def uses_format(self) -> Dict[str,List[str]]:
        return self._uses_format

    @property
    def formats(self) -> Dict[str,List[str]]:
        return self._formats

    @property
    def roots(self) -> List[FilePath]:
        return self._roots

    @property
    def root_names(self) -> List[str]:
        return self._root_names

    @property
    def keyed_roots(self) -> Dict[str,FilePath]:
        return self._keyed_roots

    @property
    def config(self) -> Config:
        return self._config

    @config.setter
    def config(self, config:Config) -> None:
        self._config = config





from cdocs.cdocs import Cdocs
import unittest

PATH = "/Users/davidkershaw/dev/cdocs/docs"
ROOTNAME = "apiui"
ROOT = PATH + "/" + ROOTNAME

class ApiUiTests(unittest.TestCase):

    noise = False
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def test_index_html(self):
        self._print(f"ApiUiTests.test_index")
        cdocs = Cdocs(ROOT)
        doc = cdocs.get_doc("/v1/index.html")
        self._print(f"ApiUiTests.test_index: doc: {doc}")




import logging
from typing import List
from cdocs.contextual_docs import DocPath, Doc, JsonDict
from cdocs.concatter import Concatter
import json

class SimpleConcatter(Concatter):

    def __init__(self, cdocs): # can't type hint cdocs because circular
        self._cdocs = cdocs

    def concat(self, paths:List[DocPath]) -> Doc:
        result = ''
        for apath in paths:
            if apath.strip() == '':
                pass
            else:
                doc = self._cdocs.get_doc(apath)
                result += self.join(result, doc)
        return Doc(result)

    def join(self, content:str, morecontent:str) -> str:
        if content is None:
            logging.warning("SimpleConcatter.join: content is None. cannot concat.")
            if morecontent is not None:
                return morecontent
            else:
                logging.warn("SimpleConcatter.join: morecontent is also None. cannot concat.")
                return ""
        if morecontent is None:
            logging.warn("SimpleConcatter.join: morecontent is None. cannot concat.")
            return content
        j1 = self._load(content)
        if j1 is None:
            return content + " " + morecontent
        j2 = self._load(morecontent)
        if j2 is None:
            return content + " " + morecontent
        newdict = {**j1, **j2}
        return json.dumps(newdict)

    def _load(self, string) -> JsonDict:
        try:
            return JsonDict(json.loads(string))
        except:
            return None



import abc
from typing import Optional, List, Dict, Any
import logging
from cdocs.config import Config
from cdocs.simple_config import SimpleConfig
from cdocs.simple_filer import SimpleFiler
from cdocs.cdocs import Cdocs
from cdocs.contextual_docs import Doc, DocPath, FilePath, JsonDict, ContextualDocs
from cdocs.multi_context_docs import MultiContextDocs
from cdocs.context_metadata import ContextMetadata

class Context(ContextualDocs, MultiContextDocs):
    """
    Context provides retrieval against an ordered set of Cdocs.
    Cdocs are internally complete. They do not share or override content.
    get_labels is the exception in that it gets the union of all labels found
    in any of the Cdocs. The value of first use of a label name wins.
    """
    def __init__(self, metadata:ContextMetadata):
        self._metadata = metadata
        self._keyed_cdocs = { k : Cdocs(v, metadata.config, self) for k,v in metadata.keyed_roots.items() }
        self._cdocs = [ v for k,v in self.keyed_cdocs.items() ]
        self._nosplitplus = None

    @property
    def cdocs(self) -> List[Cdocs]:
        return self._cdocs

    @property
    def keyed_cdocs(self) -> Dict[str, Cdocs]:
        return self._keyed_cdocs

    @property
    def metadata(self) -> ContextMetadata:
        return self._metadata

    # ==== ContextualDocs ==================

    def get_concat_doc(self, path:DocPath) -> Optional[Doc]:
        return self.get_concat_doc_from_roots(self.metadata.root_names, path)

    def get_compose_doc(self, path:DocPath) -> Optional[Doc]:
        return self.get_compose_doc_from_roots(self.metadata.root_names, path)

    def get_doc(self, path:DocPath, notfound:Optional[bool]=True, splitplus:Optional[bool]=True) -> Optional[Doc]:
        return self.get_doc_from_roots(self.metadata.root_names, path, notfound, splitplus)

    def get_labels(self, path:DocPath, recurse:Optional[bool]=True) ->  Optional[JsonDict]:
        return self.get_labels_from_roots(self.metadata.root_names, path, recurse)

    def get_tokens(self, path:DocPath, recurse:Optional[bool]=True) ->  Optional[JsonDict]:
        return self.get_tokens_from_roots(self.metadata.root_names, path, recurse)

    def list_docs(self, path:DocPath) -> List[Doc]:
        return self.list_docs_from_roots(self.metadata.root_names, path)

    # ==== MultiContextDocs ==================

    def get_filetype( self, path:DocPath) -> str:
        return SimpleFiler(self).get_filetype(path)

    def get_root_names_accepting_path( self, path:DocPath) -> List[str]:
        filetype = self.get_filetype(path)
        l = self.metadata.accepted_by.get(filetype)
        if l is None:
            l = []
        return l

    def filter_root_names_for_path(self, roots:List[str], path:DocPath) -> List[str]:
        logging.info(f"Context.filter_root_names_for_path: starting roots: {roots}, path: {path}")
        filetype = None
        hashmark = self.metadata.config.get("filenames", "hashmark", "#")
        logging.info(f"Context.filter_root_names_for_path: hashmark: {hashmark}")
        if path.find(hashmark) > -1:
            filetype = "cdocs"
        else:
            filetype = self.get_filetype(path)
        logging.info(f"Context.filter_root_names_for_path: filetype: {filetype}")
        ab = self.metadata.accepted_by
        logging.info(f"Context.filter_root_names_for_path: ab: {ab}")
        aroots = ab.get(filetype)
        if aroots is None:
            aroots = []
        logging.info(f"Context.filter_root_names_for_path: found {aroots} for filetype. filtering roots using that list.")
        filtered = [item for item in roots if item in aroots]
        if roots != filtered:
            logging.info(f"Context.filter_root_names_for_path: filtered (by accepted) {roots} to {filtered}")
        return filtered

    # TODO: this method doesn't return JsonDict. fix hint.
    def list_docs_from_roots(self, rootnames:List[str], path:DocPath) ->  Optional[JsonDict]:
        logging.info(f"Context.list_docs_from_roots: rootnames: {rootnames}, path: {path}")
        docs = []
        rootnames = self.filter_root_names_for_path(rootnames, path)
        logging.info(f"Context.list_docs_from_roots: filtered rootnames for path: {rootnames}")
        for _ in rootnames:
            cdocs = self.keyed_cdocs[_]
            logging.info(f"Context.list_docs_from_roots: cdocs: {cdocs.rootname}")
            somedocs = cdocs.list_docs(path)
            logging.info(f"Context.list_docs_from_roots: found {len(somedocs) if somedocs is not None else '0'}")
            for doc in somedocs:
                docs.append(doc)
        return docs

    def get_labels_from_roots(self, rootnames:List[str], path:DocPath, recurse:Optional[bool]=True) ->  Optional[JsonDict]:
        labels = {}
        rootnames = self.filter_root_names_for_path(rootnames, path)
        for _ in rootnames:
            cdocs = self.keyed_cdocs[_]
            labels = { **cdocs.get_labels(path, recurse), **labels }
        return labels

    def get_tokens_from_roots(self, rootnames:List[str], path:DocPath, recurse:Optional[bool]=True) ->  Optional[JsonDict]:
        tokens = {}
        rootnames = self.filter_root_names_for_path(rootnames, path)
        for _ in rootnames:
            cdocs = self.keyed_cdocs[_]
            tokens = { **cdocs.get_tokens(path, recurse), **tokens }
        return tokens

    def get_compose_doc_from_roots(self, rootnames:List[str], path:DocPath, notfound:Optional[bool]=True) -> Optional[Doc]:
        rootnames = self.filter_root_names_for_path(rootnames, path)
        for _ in rootnames:
            cdocs = self.keyed_cdocs[_]
            doc = cdocs.get_compose_doc(path)
            if doc is not None:
                return doc
        if notfound:
            return self._get_default_not_found()

    def get_concat_doc_from_roots(self, rootnames:List[str], path:DocPath, notfound:Optional[bool]=True) -> Optional[Doc]:
        rootnames = self.filter_root_names_for_path(rootnames, path)
        for _ in rootnames:
            cdocs = self.keyed_cdocs[_]
            doc = cdocs.get_concat_doc(path)
            if doc is not None:
                return doc
        if notfound:
            return self._get_default_not_found()

    def get_doc_from_roots(self, rootnames:List[str], path:DocPath, notfound:Optional[bool]=True, splitplus:Optional[bool]=True) -> Optional[Doc]:
        """
            rootnames: a list of named roots to search
            path: the docpath. may have hash and plusses
            notfound: if true, return a default notfound if no results
            splitplus: if true, plus concats can be on different roots.
                       i.e. for /x/y/z+a+b /x/y/z, /x/y/z/a, /x/y/z/b
                       can all be on different roots.
        """
        logging.info(f"Context.get_doc_from_roots: first match wins. rootnames: {rootnames}, path: {path}, notfound: {notfound}, splitplus: {splitplus}")
        plusmark = self._metadata.config.get("filenames", "plus")
        plus = path.find( plusmark )
        if plus > -1 and splitplus:
            if self._nosplitplus is None:
                nsp = self._metadata.config.get("defaults","nosplitplus", "")
                self._nosplitplus = nsp.split(',')
            if len(self._nosplitplus) > 0:
                rootnames = [name for name in rootnames if not name in self._nosplitplus]
                logging.info(f"Context.get_doc_from_roots: nsp filtered rootnames: {rootnames}")
            # split into paths and call get_doc_from_roots on each, then concat
            #
            #  /r/o/o/t.html#fish
            # needs to become /r/o/o
            #
            #  /r/o/o/t#fish
            # needs to become /r/o/o/t
            #
            #  /r/o/o/t
            # needs to become /r/o/o/t
            #
            logging.info(f"Context.get_doc_from_roots: path: {path}")
            logging.info(f"Context.get_doc_from_roots: rootnames: {rootnames}")
            logging.info(f"Context.get_doc_from_roots: notfound: {notfound}")
            logging.info(f"Context.get_doc_from_roots: splitplus: {splitplus}")
            paths = path.split(plusmark)
            logging.info(f"Context.get_doc_from_roots: paths: {paths}")
            rootpath = paths[0]
            logging.info(f"Context.get_doc_from_roots: rootpath: {rootpath}")
            hashmark = self._metadata.config.get("filenames", "hashmark")
            logging.info(f"Context.get_doc_from_roots: hashmark: {hashmark}")
            rootpath = rootpath.split( hashmark )[0]
            logging.info(f"Context.get_doc_from_roots: rootpath: {rootpath}")
            paths = [p if p.find(rootpath) > -1 else rootpath + "/" + p for p in paths]
            logging.info(f"Context.get_doc_from_roots: paths: {paths}")
            result = []
            for path in paths:
                logging.info(f"Context.get_doc_from_roots: .... next path: {path}")
                r = self.get_doc_from_roots(rootnames, path, notfound)
                if r is not None:
                    result.append(r)
            if len(result) == 0 and notfound:
                return self._get_default_not_found()
            return "".join(result)
        else:
            rootnames = self.filter_root_names_for_path(rootnames, path)
            logging.info(f"Context.get_doc_from_roots: rootnames: {rootnames} - not spliting pluses, first root locks in the pluses")
            for _ in rootnames:
                cdocs = self.keyed_cdocs[_]
                logging.info(f"Context.get_doc_from_roots: cdocs: {_} -> {cdocs.get_doc_root()}")
                doc = cdocs.get_doc(path, False)
                logging.info(f"found doc: {type(doc)}")
                if doc is not None:
                    return doc
            if notfound:
                return self._get_default_not_found()

    def _get_default_not_found(self) -> Optional[Doc]:
        notfound = self._metadata.config.get("defaults", "notfound")
        if notfound is None:
            return None
        if notfound.find("/") == -1:
            logging.error(f"Context.get_doc_from_roots: notfound {notfound} should be in form root/docpath. you should fix this.")
            return None
        root = notfound[0:notfound.index("/")]
        docpath = notfound[notfound.index("/"):]
        cdocs = self.keyed_cdocs[root]
        doc = cdocs.get_doc(docpath, False)
        if doc is None:
            logging.error(f"Context.get_doc_from_roots: notfound {notfound} in root: {root} with docpath: {docpath} is None. you should fix this.")
        return doc



from cdocs.cdocs import Cdocs, DocNotFoundException, BadDocPath
import unittest
import datetime
import time
import os

PATH:str = "/Users/davidkershaw/dev/cdocs/docs/example"

class ChangerTests(unittest.TestCase):

    noise = True
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def off(self) -> bool:
        return False

    def test_last_change(self):
        self._print(f"ChangerTests.test_last_change")
        if self.off(): return
        cdocs = Cdocs(PATH)
        cdocs.track_last_change = True
        dt = cdocs.get_last_change()
        self._print(f"Changer.test_last_change: dt: {dt}")
        lcf = cdocs._get_last_change_file_path()
        exists = os.path.exists(lcf)
        self._print(f"ChangerTests.test_last_change: lcf: {lcf}: {exists}")
        self.assertEqual(exists, True, msg=f'there must be a last change file at {lcf}')
        self.assertIsNotNone(dt, msg=f'last change cannot be None')
        time.sleep(1)

        cdocs.set_last_change()
        dt2 = cdocs.get_last_change()
        self.assertNotEqual(dt,dt2,msg=f"last change: {dt} must not equal {dt2}")
        time.sleep(1)
        cdocs = Cdocs(PATH)
        cdocs.track_last_change = True
        dt3 = cdocs.get_last_change()
        self.assertEqual(dt3,dt2,msg=f"last change: {dt3} must equal {dt2}")
        cdocs.set_last_change()
        dt4 = cdocs.get_last_change()
        self.assertNotEqual(dt4,dt3,msg=f"last change: {dt4} must not equal {dt3}")





import logging
from typing import Union
from cdocs.contextual_docs import DocPath, FilePath
from cdocs.filer import Filer

class SimpleFiler(Filer):

    def __init__(self, parent):
        # parent can be either cdocs or context, but we can't hint.
        # anyway for now it's passed pro forma.
        self._binary = ["gif","jpg","jpeg","png","pdf"]
        self._parent = parent

    def get_filetype( self, path:Union[DocPath,FilePath]) -> str:
        filetype = ''
        if path.find('/') == -1:
            filetype = 'cdocs'
        else:
            last = path.rindex('/')
            filename = path[last:]
            ext = filename.find('.')
            if ext == -1:
                filetype = 'cdocs'
            else:
                filetype = filename[ext+1:]
        return filetype

    def is_probably_not_binary(self, path:Union[DocPath,FilePath] ) -> bool:
        # todo: this method could be more sophisticated
        filetype = self.get_filetype(path)
        return filetype not in self._binary





import abc
from typing import Optional
import datetime

class Changer(metaclass=abc.ABCMeta):
    """
    Changer knows how to hold a last change date
    """

    @abc.abstractmethod
    def get_last_change(self) -> datetime:
        pass

    @abc.abstractmethod
    def set_last_change(self) -> None:
        pass

    @abc.abstractmethod
    def reset_last_change(self, dt:datetime, root:Optional[str]=None) -> None:
        pass






from cdocs.context_metadata import ContextMetadata
import unittest
import os

class MetadataTests(unittest.TestCase):

    noise = False
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def off(self) -> bool:
        return False

    def test_formats(self):
        self._print(f"MetadataTests.test_formats")
        if self.off(): return
        metadata = ContextMetadata()
        formats = metadata.formats["public"]
        self._print(f"MetadataTests.test_formats: formats are: {formats}")
        self.assertEqual(2, len(formats), msg=f'formats for public must be 2, not {len(formats)}')
        self.assertIn("xml", formats, msg=f"formats must include xml: {formats}")
        users = metadata.uses_format["xml"]
        self._print(f"MetadataTests.test_formats: users are: {users}")
        self.assertEqual(2, len(users), msg=f'users of xml must be 2, not {len(users)}')
        self.assertIn("public", users, msg=f"users of xml must include public: {users}")





from cdocs.cdocs import Cdocs, DocNotFoundException, BadDocPath
from cdocs.context import Context
from cdocs.context_metadata import ContextMetadata
import unittest
import os

class ContextTests(unittest.TestCase):

    noise = False
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def off(self) -> bool:
        return False

    def test_split_plus_concat(self):
        # tests the use of:
        #    split-plus searches
        #    non-split-plus searches
        #    split-plus searches with roots that deny split-plus
        #    implicitly, the order of multi-root searches
        # see config.ini re: order of roots under [doc] in the
        # context of split-plus searches
        docpath = "app/home/teams#delete_assignee+todo"
        metadata = ContextMetadata()
        context = Context(metadata)

        # split plus: true
        #    both roots allow split plus
        doc = context.get_doc(docpath)
        print(f"\nContextTests.test_split_plus_concat: doc 1:\n{doc}")
        self.assertIsNotNone(doc, msg=f"doc at {docpath} must not be none")
        found = doc.find("edit assignee") > -1
        found = found and doc.find("my app name: you should see") > -1
        self.assertEqual(found, True, msg=f"doc: {doc} doesn't have expected content")

        # split plus: false
        #    both roots allow split plus
        doc = context.get_doc(docpath, False, False)
        print(f"\nContextTests.test_split_plus_concat: doc 2:\n{doc}")
        found = doc.find("edit assignee") > -1
        found = found and doc.find("my app name: you should see") > -1
        self.assertEqual(found, False, msg=f"doc: {doc} doesn't have expected content")

        context._nosplitplus = ['internal','public']

        # split plus: true
        #    both roots deny split plus
        doc = context.get_doc(docpath)
        print(f"\nContextTests.test_split_plus_concat: doc 3:\n{doc}")
        self.assertIsNotNone(doc, msg=f"doc at {docpath} must not be none")
        found = doc.find("edit assignee") == -1
        found = found and doc.find("my app name: you should see") == -1
        self.assertEqual(found, True, msg=f"doc: {doc} doesn't have expected content")
        # split plus: false
        #    both roots deny split plus
        doc = context.get_doc(docpath, False, False)
        print(f"\nContextTests.test_split_plus_concat: doc 4:\n{doc}")
        found = doc.find("edit assignee") == -1
        found = found and doc.find("my app name: you should see") == -1
        self.assertEqual(found, False, msg=f"doc: {doc} doesn't have expected content")

        context._nosplitplus = ['public']

        # split plus: true
        #    public denies split plus
        doc = context.get_doc(docpath)
        print(f"\nContextTests.test_split_plus_concat: doc 5:\n{doc}")
        self.assertIsNotNone(doc, msg=f"doc at {docpath} must not be none")
        found = doc.find("edit assignee") > -1
        found = found and doc.find("my app name: you should see") == -1
        self.assertEqual(found, True, msg=f"doc: {doc} doesn't have expected content")
        # split plus: false
        #    public denies split plus
        doc = context.get_doc(docpath, False, False)
        print(f"\nContextTests.test_split_plus_concat: doc 6:\n{doc}")
        found = doc.find("edit assignee") == -1
        found = found and doc.find("my app name: you should see") > -1
        self.assertEqual(found, True, msg=f"doc: {doc} doesn't have expected content")

        context._nosplitplus = ['internal']

        # split plus: true
        #    internal denies split plus
        doc = context.get_doc(docpath)
        print(f"\nContextTests.test_split_plus_concat: doc 7:\n{doc}")
        self.assertIsNotNone(doc, msg=f"doc at {docpath} must not be none")
        found = doc.find("edit assignee") == -1
        found = found and doc.find("my app name: you should see") > -1
        self.assertEqual(found, True, msg=f"doc: {doc} doesn't have expected content")

        # split plus: false
        #    internal denies split plus
        doc = context.get_doc(docpath, False, False)
        print(f"\nContextTests.test_split_plus_concat: doc 8:\n{doc}")
        found = doc.find("edit assignee") == -1
        found = found and doc.find("my app name: you should see") > -1
        self.assertEqual(found, True, msg=f"doc: {doc} doesn't have expected content")

    def test_roots(self):
        self._print("ContextTests.test_roots")
        if self.off(): return
        metadata = ContextMetadata()
        roots = metadata.roots
        self.assertEqual( len(roots), 5, msg="must be 4 roots" )

    def test_accepts(self):
        self._print("ContextTests.test_accepts")
        if self.off(): return
        metadata = ContextMetadata()
        accepts = metadata.accepts
        if self.noise: self._print(f"accepts: {accepts}")
        self.assertIsNotNone( accepts, msg="accepts must not be None")
        self.assertEqual(len(accepts), 5, msg=f"must be 4 keys in accepts, not {accepts}")
        self.assertIsNotNone( accepts["images"], msg="accepts must have an 'images' key")
        images = accepts["images"]
        self.assertEqual(len(images), 4, msg="must be 4 items in images")
        self.assertIn("gif", images, msg=f"images must include item 'gif'")
        self.assertIn("png", images, msg=f"images must include item 'png'")
        self.assertIn("jpg", images, msg=f"images must include item 'jpg'")
        self.assertIn("jpeg", images, msg=f"images must include item 'jpeg'")

    def test_accepted_by(self):
        self._print("ContextTests.test_accepted_by")
        if self.off(): return
        metadata = ContextMetadata()
        accceptedby = metadata.accepted_by
        self._print(f"accceptedby: {accceptedby}")
        self.assertIsNotNone( accceptedby, msg="accceptedby must not be None")
        self.assertEqual(len(accceptedby), 7, msg=f"must be 7 keys in accceptedby, not {accceptedby}")
        self.assertIsNotNone( accceptedby["cdocs"], msg="accepts must have a 'cdocs' key")
        cdocs = accceptedby["cdocs"]
        self.assertEqual(len(cdocs), 2, msg="must be 2 items in cdocs")
        self.assertIn("public", cdocs, msg=f"cdocs must include item 'public'")
        self.assertIn("internal", cdocs, msg=f"cdocs must include item 'internal'")
        gif = accceptedby["gif"]
        self.assertIn("images", gif, msg=f"gif must include item 'images'")

    def test_get_filetype(self):
        self._print("ContextTests.test_get_filetype")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        atype = context.get_filetype("/x/y/z.gif")
        self._print(f"type for /x/y/z.gif: {atype}")
        self.assertEqual(atype, "gif", msg="type must be gif")
        atype = context.get_filetype("/x/y/z")
        self._print(f"type for /x/y/z: {atype}")
        self.assertEqual(atype, 'cdocs', msg="type must be cdocs")

    def test_filter_root_names_for_path(self):
        self._print("ContextTests.test_filter_root_names_for_path")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        roots = ["images", "fish", "public"]
        rs = context.filter_root_names_for_path(roots, "/x/y/z")
        self._print(f"filtered roots: {rs}")
        self.assertIsNotNone( rs, msg="filtered roots must not be None")
        self.assertEqual(len(rs), 1, msg="must be 1 filtered root")
        self.assertEqual(rs, ["public"], msg="filtered roots must be ['public']")


    def test_get_root_names_accepting_path(self):
        self._print("ContextTests.test_get_root_names_accepting_path")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        gifs = context.get_root_names_accepting_path("/x/y/z.gif")
        self._print(f"roots for /x/y/z.gif: {gifs}")
        self.assertIsNotNone( gifs, msg="gifs must not be None")
        self.assertEqual(len(gifs), 1, msg="must be 1 items in gifs")
        cdocs = context.get_root_names_accepting_path("/x/y/z")
        self._print(f"roots for /x/y/z: {cdocs}")
        self.assertIsNotNone( cdocs, msg="cdocs must not be None")
        self.assertEqual(len(cdocs), 2, msg="must be 2 items in cdocs")

    def test_create_context(self):
        self._print("ContextTests.test_context")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        cdocs = context.cdocs
        self.assertEqual( len(cdocs), 5, msg=f"must be 4 cdocs, not {len(cdocs)} from {cdocs}" )

    def test_get_known_type_doc(self):
        self._print(f"ContextTests.test_get_known_type_doc")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        docpath = "/app/home/teams/3-copy.png"
        f = context.get_doc(docpath)
        self._print(f"the image doc {type(f)} ")
        self.assertIsNotNone(f, msg=f"doc at {docpath} must not be none")
        self.assertEqual( type(f).__name__, 'bytes', msg=f"type must be 'bytes'")

    def test_get_doc(self) :
        self._print(f"ContextTests.test_get_doc")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        docpath = "/app/home/teams/todos/assignee"
        txt = context.get_doc(docpath)
        self.assertIsNotNone(txt, msg=f"doc at {docpath} must not be none")
        self._print(f"test_get_doc: doing get_docs. the doc txt is: {txt}")
        point = txt.find("assignee in company starstruck!")
        self.assertNotEqual(-1,point, msg="must include 'assignee in company starstruck!'")
        point = txt.find("my app name: you should see: my app's name is fruit")
        self.assertNotEqual(-1,point, msg="must include 'my app name: you should see: my app's name is fruit'")

    def test_get_doc_just_checking(self) :
        self._print(f"ContextTests.test_get_doc_just_checking")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        docpath = "/app/home/teams*delete_assignee"
        txt = context.get_doc(docpath)
        # this doc could be a not-found doc. for what I need today that's fine.
        self.assertIsNotNone(txt, msg=f"doc at {docpath} must not be none")
        self._print(f"ContextTests.test_get_doc_just_checking: doing get_docs. the doc txt is: {txt}")

    def test_get_labels(self):
        self._print(f"ContextTests.test_get_labels")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee"
        #docpath = "/v1/test/fish"
        metadata = ContextMetadata()
        context = Context(metadata)
        labels = context.get_labels(docpath)
        print(f"test_get_labels: the labels are: {labels}")
        self.assertIn("app", labels, msg=f"labels at {docpath} must include key 'app'")
        self.assertIn("team", labels, msg=f"labels at {docpath} must include key 'team'")
        self.assertIn("my_app_name", labels, msg=f"labels at {docpath} must include key 'my_app_name'")
        self.assertEqual("my app's name is fruit", labels["my_app_name"], msg=f"label my app name must == my app's name is fruit")
        self.assertEqual("starstruck", labels["company"], msg=f"label company must == starstruck")

    def test_get_doc_from_wrong_root(self) :
        self._print(f"ContextTests.test_get_doc_from_wrong_root")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        docpath = "/app/home/teams/todos/assignee"
        txt = context.get_doc_from_roots(["images"], docpath, False)
        self.assertIsNone(txt, msg=f"doc at {docpath} in root 'images' must be none")

    def test_get_doc_from_right_root(self) :
        self._print(f"ContextTests.test_get_doc_from_right_root")
        if self.off(): return
        metadata = ContextMetadata()
        context = Context(metadata)
        docpath = "/app/home/teams/todos/assignee"
        txt = context.get_doc_from_roots(["public"], docpath)
        self.assertIsNotNone(txt, msg=f"doc at {docpath} in root 'public' must not be none")


    def test_get_compose_doc_with_roots(self):
        self._print(f"ContextTests.test_get_compose_doc_with_roots")
        if self.off(): return
        docpath = "/app/home/teams"
        metadata = ContextMetadata()
        context = Context(metadata)
        doc = context.get_doc_from_roots(["internal"], docpath, False)
        self._print(f"test_get_compose_doc_with_roots: doc from 'internal': {doc}")
        assignee = doc.find('assignee in company starstruck!')
        self.assertNotEqual(-1, assignee, msg=f'{docpath} must include "assignee in company starstruck!" in {doc}')
        edit = doc.find('edit assignee')
        self.assertNotEqual(-1, edit, msg=f'{docpath} must include "edit assignee" in {doc}')
        doc = context.get_doc_from_roots(["public"], docpath)
        self._print(f"test_get_compose_doc_with_roots: doc from 'public': {doc}")
        notfound = doc.find("Not found!")
        self.assertNotEqual(notfound, -1, msg=f'doc at {docpath} must include "Not found!"')






import os
import logging
from cdocs.contextual_docs import DocPath, FilePath
from cdocs.simple_config import SimpleConfig
from cdocs.pather import Pather
from typing import Optional

class SimplePather(Pather):

    def __init__(self, metadata, cdocs): #docspath:FilePath, config:FilePath=None):
        logging.info(f"SimplePather.__init__: starting with metadata: {metadata} and cdocs: {cdocs}")
        cfg = metadata.config
        logging.info(f"SimplePather.__init__: cfg: {cfg}")
        self._hashmark:str  = cfg.get("filenames", "hashmark", "#")
        logging.info(f"SimplePather.__init__: hashmark: {self._hashmark}")
        self._docs_path:str = cdocs.get_doc_root()
        self._rootname = cfg.get_matching_key_for_value("docs", cdocs.get_doc_root())
        logging.info(f"SimplePather.__init__: docspath: {cdocs.get_doc_root()}, rootname: {self._rootname}")
        ext = cfg.get("defaults", "ext", "xml")
        ext = cfg.get("formats", self._rootname, ext)
        logging.info(f"SinplePather.__init__: ext: {ext}")
        if ext.find(",") > -1:
            self._exts = ext.split(",")
        else:
            self._exts = [ext]
        logging.info(f"SimplePather.__init__: exts: {self._exts}")

    def get_full_file_path(self, path:DocPath) -> FilePath:
        return self.get_full_file_path_for_root(path, self._docs_path)

    def get_full_file_path_for_root(self, path:DocPath, root:FilePath) -> FilePath:
        logging.info(f"SimplePather.get_full_file_path_for_root: path: {path}, root: {root}")
        path = path.strip('/\\')
        if path == '':
            logging.info(f"SimplePather.get_full_file_path_for_root: path points to root. returning root.")
            return root
        logging.info(f"SimplePather.get_full_file_path_for_root: getting filename for {path}")
        filename = self.get_filename(path)
        logging.info(f"SimplePather.get_full_file_path_for_root: path: {path}, filename: {filename}, root: {root}")
        if filename is None:
            pass
        else:
            path = path[0:path.find(self._hashmark)]
        #
        # why do we change out root here?  seems to do no harm, but...
        #
        #logging.info(f"SimplePather.get_full_file_path_for_root: root: {root}, _docs_path: {self._docs_path}")
        #root = self._docs_path
        #
        # if path were None we'd never get here!
        #
        #if path is None:
        #    logging.error(f"SimplePather.get_full_file_path_for_root: path is None")
        path = os.path.join(root, path)
        apath = path
        logging.info(f"SimplePather.get_full_file_path_for_root: joined root: {root} with path to get: {apath}")
        i = path.find(".")
        logging.info(f"SimplePather.get_full_file_path_for_root: index of '.': {i}")
        if filename is None and i <= len(root):
            logging.info(f"SimplePather.get_full_file_path_for_root: filename is None and no '.'")
            apath = self._find_path(apath)
        elif filename is None:
            logging.info(f"SimplePather.get_full_file_path_for_root: filename is None and there is a '.'")
            pass
        else:
            logging.info(f"SimplePather.get_full_file_path_for_root: apath: {apath}, last char: {apath[-1:]}, filename: {filename}")
            apath = apath + ('' if apath[-1:] == '/' else os.path.sep )+ filename
            apath = self._find_path(apath)
        if apath is None:
            logging.info(f"SimplePather.get_full_file_path_for_root: apath is None! from: {self._rootname}->{apath}")
        return FilePath(apath)

    #
    # 1. doesn't seem to look specifically for cdocs paths?
    # 2. see notes below
    #
    def _find_path(self, path) -> Optional[FilePath]:
        logging.info(f"SimplePather._find_path: starting path: {path}")
        logging.info(f"SimplePather._find_path: looking for path using exts in {self._exts}")
        for ext in self._exts:
            apath = None
            anext = path[-1*len(ext):]
            logging.info(f"SimplePather._find_path: checking if path's anext: {anext} matches ext: {ext}")
            if anext == ext:
                apath = path
            else:
                apath = path + "." + ext
            logging.info(f"SimplePather._find_path: checking for a simple file: apath: {apath}")
            if os.path.exists(apath):
                logging.info(f"SimplePather._find_path: apath exists. returning: {apath}")
                return apath
        if len(self._exts) == 1:
            logging.info(f"SimplePather._find_path: no file. just one ext. we assume name + '.' + ext")
            return path + "." + self._exts[0]
        elif len(self._exts) > 1:
            #
            # if any of the paths exists is knowable. shouldn't we check?
            #
            logging.info(f"SimplePather._find_path: on file. there are {len(self._exts)} exts, so we guess the first one, but check for actual files -- keep in mind, if the file exists or not is not a pather's problem.")
            presumedpath = path + "." + self._exts[0]
            logging.info(f"SimplePather._find_path: presumed path is {presumedpath}. now checking alternatives.")
            for _ in self._exts:
                apath = path + "." + _
                logging.info(f"SimplePather._find_path: checking if {apath} exists")
                if os.path.exists(apath):
                    logging.info(f"SimplePather._find_path: {apath} exists! returning it.")
                    return apath
            logging.info(f"SimplePather._find_path: no path exists. returning presumed path: {presumedpath}")
            return presumedpath
        else:
            logging.warning("SimplePather._find_path: no exts! you should fix this.")
        return None

    def get_filename(self, path:str) -> Optional[str]:
        logging.info(f"SimplePather.get_filename: path: {path}. if a hashmark is found it marks a filename. if no hashmark we return none.")
        filename = None
        hashmark = path.find(self._hashmark)
        if hashmark > -1:
            filename = path[hashmark+1:]
        logging.info(f"SimplePather.get_filename: returning: {filename if filename is not None else 'intentionally returning no filename'}")
        return filename




import abc
from typing import Optional, List, Tuple
from cdocs.physical import Physical

class ConfigException(Exception):
    pass


class Config(Physical, metaclass=abc.ABCMeta):
    """
    Config is a sectioned key-value store interface that follows configparser
    """

    @abc.abstractmethod
    def get_config_path(self):
        pass

    @abc.abstractmethod
    def get(self, group:str, name:str, default:Optional[str]=None) -> str:
        pass

    @abc.abstractmethod
    def get_items(self, group:str, exceptnot:List[str]=None) -> List[Tuple[str, str]]:
        pass

    @abc.abstractmethod
    def get_matching_key_for_value(self, group:str, value:str) -> Optional[str]:
        pass



from configparser import ConfigParser, NoSectionError
import os
from typing import Optional, List, Tuple
import logging
from cdocs.config import Config

class SimpleConfig(Config):

    def __init__(self, path:Optional[str]=None):
        super().__init__()
        if path is None:
            logging.debug(f"cdocs Config.__init__ without config path. using default: {os.getcwd()}/config/config.ini")
            self._path = 'config/config.ini'
        else:
            self._path = path
        self._parser = ConfigParser()
        self._parser.read(self._path)

    @property
    def parser(self) -> ConfigParser:
        return self._parser

    def get_config_path(self):
        return self._path

    # deprecated: just use get
#    def get_with_default(self, group, name, default:Optional[str]=None) -> str:
#        val = self.get(group, name)
#       if val is None:
#            val = default
#        return val

    def get(self, group:str, name:str, default:Optional[str]=None) -> str:
        #return self.get_with_default(group,name,default)
        logging.info(f"SimpleConfig.get: {group}, {name}, {default}")
        try:
            val = self._parser.get(group, name)
            logging.info(f"SimpleConfig.get: val: {val}")
            if val is None:
                return default
            return val
        except Exception as e:
            logging.info(f"Cdocs Config.get: unable to get [{group}][{name}]: {e}. returning default: {default}.")
            return default

    def get_items(self, group:str, exceptnot:List[str]=None) -> List[Tuple[str, str]]:
        items = None
        try:
            items = self._parser.items(group)
        except (KeyError, NoSectionError):
            logging.info("Cdocs Config.get_items: no such group {group}. returning [].")
            items = []
        if exceptnot is not None:
            items = [ _ for _ in items if _[0] not in exceptnot]
        return items

    def get_matching_key_for_value(self, group:str, value:str) -> Optional[str]:
        logging.info(f"SimpleConfig.get_matching_key_for_value: group: {group}, value: {value}")
        items = self.get_items(group)
        logging.info(f"SimpleConfit.get_matching_key_for_value: items: {items}")
        for item in items:
            logging.info(f"SimpleConfit.get_matching_key_for_value: {item[1]} == {value}: {item[1] == value}")
            if item[1] == value:
                return item[0]
        return None





import abc
from cdocs.contextual_docs import DocPath
from typing import Optional, Dict

class Transformer(metaclass=abc.ABCMeta):
    """
    Transformer knows how to convert a string to its final form
    """

    @abc.abstractmethod
    def transform(self, content:str, path:DocPath=None, \
                   tokens:Optional[Dict[str,str]]=None, transform_labels=True) -> str:
        pass




from cdocs.cdocs import Cdocs, DocNotFoundException, BadDocPath
from cdocs.context import Context
from cdocs.simple_config import SimpleConfig
from cdocs.context_metadata import ContextMetadata
import unittest
import os
import logging
from typing import Optional

class Repl(object):

    def __init__(self):
        self._config = None
        self._metadata = None
        self._context = None
        self._continue = True
        self._debug = False
        self._commands = {}
        self._logger = logging.getLogger('')
        self._add_commands()

    def _add_commands(self):
        self.commands["quit"] = self.quit
        self.commands["read"] = self.read
        self.commands["list"] = self.list
        self.commands["roots"] = self.roots
        self.commands["labels"] = self.labels
        self.commands["tokens"] = self.tokens
        self.commands["debug"] = self.debug
        self.commands["help"] = self.help
        self.commands["?"] = self.help

    @property
    def commands(self):
        return self._commands

    @commands.setter
    def commands(self, adict):
        self._commands = adict

    def ask_debug(self):
        d = input("want to debug during load? (y/n) ")
        if (d=='y'):
            self.debug()

    def ask_config(self):
        d = input("want to use your own config? (y/n) ")
        if (d=='y'):
            return input("what path? ")
        return None

    def setup(self, configpath:Optional[str]=None, askdebug:Optional[bool]=True):
        self.ask_debug()
        configpath = self.ask_config()
        metadata = None
        self._config = SimpleConfig(configpath)
        self._metadata = ContextMetadata(self._config)
        self._context = Context(self._metadata)

    def loop(self):
        print("\n")
        while self._continue:
            self._one_loop()

    def _one_loop(self) -> bool:
        cmd = input("cmd: ")
        self.do_cmd(cmd)

    def do_cmd(self, cmd):
        print(f"do_cmd: {cmd}")
        callme = self.commands.get(cmd)
        if callme is None:
            print(f"not sure what {cmd} means")
        else:
            callme()

    def debug(self):
        if self._debug:
            self._logger.setLevel(level=logging.WARN)
            self._logger.warning("Set level to WARN")
            self._debug = False
        else:
            self._logger.setLevel(level=logging.DEBUG)
            self._logger.debug("Set level to DEBUG")
            self._debug = True

    def help(self):
        print("\nHelp:")
        for k,v in self.commands.items():
            print(f"   {k}")
        return True

    def read(self):
        roots = self._get_roots()
        docpath = input("docpath: ")
        doc = ""
        try:
            if len(roots) >= 1:
                doc = self._context.get_doc_from_roots(roots, docpath)
            else:
                doc = self._context.get_doc(docpath)
            print("\ndoc: ")
            print(f"{doc}")
        except BadDocPath as e:
            print(f"Error: {e}")
        return True

    def labels(self):
        roots = self._get_roots()
        docpath = input("docpath: ")
        labels = ""
        try:
            if len(roots) >= 1:
                labels = self._context.get_labels_from_roots(roots, docpath)
            else:
                labels = self._context.get_labels(docpath)
            print("\nlabels: ")
            print(f"{labels}")
        except BadDocPath as e:
            print(f"Error: {e}")
        return True

    def tokens(self):
        roots = self._get_roots()
        docpath = input("docpath: ")
        labels = ""
        try:
            if len(roots) >= 1:
                labels = self._context.get_tokens_from_roots(roots, docpath)
            else:
                tokens = self._context.get_tokens(docpath)
            print("\ntokens: ")
            print(f"{tokens}")
        except BadDocPath as e:
            print(f"Error: {e}")
        return True

    def _get_roots(self):
        roots = input("which roots (Csv or return for all): ")
        print("roots are: " + roots)
        roots = [] if roots == "" else roots.split(",")
        print(f"roots are: {roots}")
        return roots

    def list(self):
        roots = self._get_roots()
        docpath = input("docpath: ")
        docs = []
        if len(roots) >= 1:
            docs = self._context.list_docs_from_roots(roots, docpath)
        else:
            docs = self._context.list_docs(docpath)
        print("\ndocs: ")
        for doc in docs:
            print(f"   {doc}")
        return True

    def roots(self):
        roots = self._config.get_items("docs")
        print("\nroots:")
        for root in roots:
            print(f"   {root}")
        return True

    def quit(self):
        self._continue = False
        return True


if __name__ == "__main__":
    repl = Repl()
    repl.setup()
    repl.loop()





from cdocs.cdocs import Cdocs, DocNotFoundException, BadDocPath
import unittest
import os
import logging

PATH:str = "/Users/davidkershaw/dev/cdocs/docs/example"
PATH2:str = "/Users/davidkershaw/dev/cdocs/docs/internal"

class CdocsSchemeTests(unittest.TestCase):

    logger = logging.getLogger('')
    def _debug(self):
        self.logger.setLevel(level=logging.DEBUG)
        self.logger.debug("SET THE LEVEL TO DEBUG")

    noise = True
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def off(self) -> bool:
        return False


    def test_finding_doc_at_root_using_cdoc(self):
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc")
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: THIS TEST IS NOT YET DETERMINISTIC.")
        #
        # we have:
        #   /404x.html
        #   /404
        #   /404a.xml
        #
        # which is the concept?
        # concepts are:
        #    directory names
        #    any file without an extension
        #    any file with an extension in the exts array
        # everything else must be addressed as directory#filename
        # so:
        #    /404 is a concept you get with /404
        #    /404a.xml is a concept (xml is in the exts array) you get with /404a
        #    /404x.html is a file you get using /#404x.html
        #
        #
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: NOT A DETERMINISTIC TEST!")
        cdocs = Cdocs(PATH)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc. cdocs: {cdocs}")
        self._print("")
        docpath1 = "/404x.html"  # should be not found because not in root with accepts html
        docpath2 = "/#404x.html" # same
        docpath3 = "/404x"       # not found. this is correct, 404x is html
        docpath4 = "/#404x"      # not found. this is correct, 404x is html
        docpath5 = "/404a"       # found because 404a.xml exists
        docpath6 = "/#404a"      # found because 404a.xml exists. concepts can also be addressed as supporting files.
        docpath7 = "/404a.xml"   # not found because its a full file path not cdocs path
        docpath8 = "/#404a.xml"  # found because can find a full filename on hashmark with cdocs in accepts. the accepts needs to have 'cdocs' because the hashmark is cdocs.
        docpath9 = "/404"        # not found. this is right. cdocs docpath may not have an extension, but files must have extensions.
        docpath10 = "/#404"      # not found. this is right, there is no 404.xml or 404.json.


        doc = cdocs.get_doc(docpath1)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath1}: doc: {doc}")

        doc = cdocs.get_doc(docpath2)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath2}: doc: {doc}")

        doc = cdocs.get_doc(docpath3)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath3}: doc: {doc}")

        doc = cdocs.get_doc(docpath4)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath4}: doc: {doc}")

        doc = cdocs.get_doc(docpath5)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath5}: doc: {doc}")

        doc = cdocs.get_doc(docpath6)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath6}: doc: {doc}")

        doc = cdocs.get_doc(docpath7)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath7}: doc: {doc}")

        doc = cdocs.get_doc(docpath8)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath8}: doc: {doc}")

        doc = cdocs.get_doc(docpath9)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath9}: doc: {doc}")

        doc = cdocs.get_doc(docpath10)
        self._print(f"CdocsSchemeTests.test_finding_doc_at_root_using_cdoc: docpath: {docpath10}: doc: {doc}")

        #self.assertIsNotNone(doc, msg=f"docpath: {docpath} must not be None")
        #found = doc.find("REALLY") >= 0
        #self.assertEqual(found, True, msg=f"doc at {docpath} must include 'REALLY'")









import abc
from typing import List
from cdocs.contextual_docs import Doc, DocPath

class Lister(metaclass=abc.ABCMeta):
    """
    Lister knows how to list the Doc below a docpath. It doesn't include
    the doc pointed to by the docpath. E.g. for /x/y/z it will return
    [/x/y/z/a.xml, /x/y/z/b.xml] (actually 'a.xml' and 'b.xml') but it
    won't include /x/y/z.xml.
    """

    @abc.abstractmethod
    def list_docs(self, path:DocPath) -> List[Doc]:
        pass





import abc
import os.path
from .contextual_docs import Doc, FilePath
from cdocs.reader import Reader
import logging
import traceback

class SimpleReader(Reader):

    def __init__(self, cdocs):
        self._cdocs = cdocs

    def read(self, filepath:FilePath) -> Doc:
        try:
            with open(filepath, 'rb') as f:
                doc = Doc(f.read())
                #print(f"SimpleReader.read: doc is {doc} of type {type(doc)}")
                return doc
        except FileNotFoundError as fnfe:
            logging.warning(f'SimpleReader.read: cannot read: {fnfe}')
            # raise error?
            return None


    def is_available(self, filepath:FilePath) -> bool:
        if filepath is None:
            logging.warning(f"SimpleReader.is_available: filepath is None")
            #traceback.print_stack(limit=7)
            return False
        return os.path.isfile(filepath)





from cdocs.pather import Pather
from cdocs.reader import Reader
from cdocs.finder import Finder

class Physical(object):
    """
    Physical (along with Config, a subclass of Physical) knows
    which classes should be used to find and read files. The
    default reader, finder, pather and config classes work from
    the local filesystem.
    """

    def __init__(self):
        self._reader:Reader = None
        self._finder:Finder = None
        self._pather:Pather = None

    @property
    def reader(self) -> Reader:
        return self._reader

    @reader.setter
    def reader(self, reader:Reader) -> None:
        self._reader = reader

    @property
    def finder(self) -> Finder:
        return self._finder

    @finder.setter
    def finder(self, finder:Finder) -> None:
        self._finder = finder

    @property
    def pather(self) -> Pather:
        return self._pather

    @pather.setter
    def pather(self, pather:Pather) -> None:
        self._pather = pather





import abc
import os
from cdocs.contextual_docs import DocPath, FilePath
from typing import Optional

class Pather(metaclass=abc.ABCMeta):
    """
    Pather handles converting DocPath to FilePath. In other words, it
    knows how to take an abstract Cdocs path and turn it into a
    physical path, by default in the local filesystem.
    """

    @abc.abstractmethod
    def get_full_file_path(self, path:DocPath) -> FilePath:
        pass

    @abc.abstractmethod
    def get_full_file_path_for_root(self, path:DocPath, root:FilePath) -> FilePath:
        pass

    @abc.abstractmethod
    def get_filename(self, path:str) -> Optional[str]:
        pass




import abc
from cdocs.contextual_docs import Doc, DocPath, FilePath


class Reader(metaclass=abc.ABCMeta):
    """
    Reader knows how to get bytes pointed to by a FilePath. By
    default the FilePath will be on the local filesystem.
    """

    @abc.abstractmethod
    def is_available(self, filepath:FilePath) -> bool:
        pass

    @abc.abstractmethod
    def read(self, filepath:FilePath) -> Doc:
        pass





import abc
from cdocs.contextual_docs import DocPath, JsonDict
from typing import Optional


class Finder(metaclass=abc.ABCMeta):
    """
    Finder knows how to collect and aggregate JsonDict. by default
    it finds them in the local filesystem.
    """

    def find_tokens(self, path:DocPath=None, filename:str="tokens.json", recurse:Optional[bool]=True) -> JsonDict:
        pass




import logging
from typing import Optional, Dict
from jinja2 import Template
from cdocs.contextual_docs import DocPath, FilePath, JsonDict
from cdocs.simple_config import SimpleConfig
from cdocs.transformer import Transformer
from cdocs.contextual_docs import DocPath
import inflect

class SimpleTransformer(Transformer):

    def __init__(self, cdocs): # can't type hint cdocs because circular
        self._cdocs = cdocs
        self._engine = inflect.engine()

    def _plural(self, word):
        return self._engine.plural(word)

    def _cap(self, word):
        return word.capitalize()

    def _article(self, word):
        return self._engine.a(word)

    def transform(self, content:str, path:DocPath=None, \
                   tokens:Optional[Dict[str,str]]=None, transform_labels=True) -> str:
        if content is None:
            logging.info("SimpleTransformer.transform: cannot transform None. returning ''")
            return None
        if path is None:
            raise BadDocPath("you must provide the DocPath")
        filetype = self._cdocs.filer.get_filetype(path)
        # more filetypes could go here, but for now this is good.
        # todo: make this list a config option?
        if filetype in ['html','concat','cdocs','xml','md','txt','xhtml','yaml','json','js']:
            if tokens is None:
                tokens:JsonDict = self._cdocs.get_tokens(path)
            if path is not None and transform_labels:
                tokens = self._cdocs._add_labels_to_tokens(path, tokens)
            tokens["get_doc"] = self._cdocs.get_doc
            tokens["get_compose_doc"] = self._cdocs.get_compose_doc
            tokens["get_concat_doc"] = self._cdocs.get_concat_doc
            tokens["plural"] = self._plural
            tokens["cap"] = self._cap
            tokens["article"] = self._article
            tokens["docroot"] = self._cdocs.get_doc_root()
            if self._cdocs.context is not None:
                tokens["get_concat_doc_from_roots"] = self._cdocs.context.get_concat_doc_from_roots
                tokens["get_compose_doc_from_roots"] = self._cdocs.context.get_compose_doc_from_roots
                tokens["get_doc_from_roots"] = self._cdocs.context.get_doc_from_roots
                tokens["get_labels_from_roots"] = self._cdocs.context.get_labels_from_roots
                logging.info("SimpleTransformer.transform: added multi root methods on context to template tokens")
            try:
                template = Template(content)
                content = template.render(tokens)
            except Exception as e:
                logging.info(f"SimpleTransformer.transform: couldn't transform content: {e}")
        return content





from cdocs.cdocs import Cdocs, DocNotFoundException, BadDocPath
import unittest
import os
import logging

PATH:str = "/Users/davidkershaw/dev/cdocs/docs/example"
PATH2:str = "/Users/davidkershaw/dev/cdocs/docs/internal"
JSON:str = "/Users/davidkershaw/dev/cdocs/docs/json"
APIUI:str = "/Users/davidkershaw/dev/cdocs/docs/apiui"

class CdocsTests(unittest.TestCase):

    logger = logging.getLogger('')
    def _debug(self):
        self.logger.setLevel(level=logging.DEBUG)
        self.logger.debug("SET THE LEVEL TO DEBUG")


    noise = True
    def _print(self, text:str) -> None:
        if self.noise:
            print(text)

    def off(self) -> bool:
        return False

    def test_notfounds(self):
        self._print(f"CdocsTests.test_notfounds")
        if self.off(): return
        cdocs1 = Cdocs(PATH)
        doc = cdocs1.get_404()
        self._print(f"test_notfounds: for {cdocs1}: {doc}")
        found = doc.find("NO")
        self.assertNotEqual(found, -1, msg=f"not found in {cdocs1.rootname} must include 'NO'")
        cdocs2 = Cdocs(PATH2)
        doc = cdocs2.get_404()
        self._print(f"test_notfounds: for {cdocs2}: {doc}")
        found = doc.find("TRY")
        self.assertNotEqual(found, -1, msg=f"not found in {cdocs2.rootname} must include 'NO'")

    def test_get_doc_root_path(self):
        self._print(f"CdocsTests.test_get_doc_root_path")
        if self.off(): return
        cdocs = Cdocs(PATH)
        path = cdocs.get_doc_root()
        self._print(f"test_get_doc_root_path: docs root is {path}")
        self.assertTrue( os.path.exists(path), msg="docs root must exsit" )
        self.assertEqual(path, PATH)

    def test_bad_path_arg(self):
        self._print(f"CdocsTests.test_bad_path_arg")
        if self.off(): return
        cdocs = Cdocs(PATH)
        self._print("test_bad_path_arg: checking for exceptions")
        with self.assertRaises(DocNotFoundException):
            cdocs.get_doc(None)
        with self.assertRaises(BadDocPath):
            cdocs.get_doc("test.xml")

    def test_get_doc(self) :
        self._print(f"CdocsTests.test_get_doc")
        if self.off(): return
        #self._debug()
        docpath = "/app/home/teams/todos/assignee"
        cdocs = Cdocs(PATH)
        txt = cdocs.get_doc(docpath)
        self.assertIsNotNone(txt, msg=f"doc at {docpath} must not be none")
        self._print(f"test_get_doc: doing get_docs. the doc txt is: {txt}")
        point = txt.find("assignee in company starstruck!")
        self.assertNotEqual(-1,point, msg=f"txt: {txt} must include 'assignee in company starstruck!'")
        point = txt.find("my app name: you should see: my app's name is fruit")
        self.assertNotEqual(-1,point, msg="must include 'my app name: you should see: my app's name is fruit'")

    def test_get_doc_not_found(self) :
        self._print(f"CdocsTests.test_get_doc_not_found")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee/fish"
        cdocs = Cdocs(PATH)
        txt = cdocs.get_doc(docpath)
        self._print(f"test_get_doc_not_found: the doc found is: {txt}")
        self.assertIsNotNone(txt, msg=f"doc at {docpath} must not be none")
        self._print(f"test_get_doc_not_found: doing get_docs. the doc txt is: {txt}")
        point = txt.find("NO")
        self.assertNotEqual(-1,point, msg=f"txt: {txt} must include 'Not found!'")


    def test_add_labels_to_tokens(self):
        self._print(f"CdocsTests.test_add_labels_to_tokens")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee#new_assignee"
        cdocs = Cdocs(PATH)
        tokens = {}
        tokens = cdocs._add_labels_to_tokens(docpath, tokens)
        self._print(f"test_add_labels_to_tokens: tokens are {tokens}.")
        self.assertIn("label__my_app_name", tokens, msg=f"must include labels_my_app_name")

    def test_get_plus_paths(self):
        self._print(f"CdocsTests.test_get_plus_paths")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee+edit_assignee"
        cdocs = Cdocs(PATH)
        paths = cdocs._get_plus_paths(docpath)
        self._print(f"test_get_plus_paths: plus paths: {paths}")
        self.assertIn("/app/home/teams/todos/assignee#edit_assignee", paths, msg=f"1. plus paths must include edit_assignee")
        docpath = "/app/home/teams/todos/assignee#new_assignee+edit_assignee"
        paths = cdocs._get_plus_paths(docpath)
        self._print(f"plus paths: {paths}")
        self.assertIn("/app/home/teams/todos/assignee#edit_assignee", paths, msg=f"2. plus paths must include edit_assignee")
        docpath = "/app/home/teams/todos/assignee+new_assignee+edit_assignee"
        paths = cdocs._get_plus_paths(docpath)
        self._print(f"plus paths: {paths}")
        self.assertIn("/app/home/teams/todos/assignee#edit_assignee", paths, msg=f"3. plus paths must include edit_assignee")
        self.assertIn("/app/home/teams/todos/assignee#new_assignee", paths, msg=f"4. plus paths must include new_assignee")

    def test_get_doc_with_plus_path(self):
        self._print(f"CdocsTests.test_get_doc_with_plus_path")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee+new_assignee+edit_assignee"
        cdocs = Cdocs(PATH)
        doc = cdocs.get_doc(docpath)
        self._print(f"test_get_doc_with_plus_path: content of {docpath} is: {doc}")
        na = doc.find("new assignee")
        ea = doc.find("edit assignee")
        a = doc.find("assignee in company")
        self.assertNotEqual(-1, na, msg=f'must include "new assignee" in {doc}')
        self.assertNotEqual(-1, ea, msg=f'must include "edit assignee" in {doc}')
        self.assertNotEqual(-1, a, msg=f'must include "assignee in company" in {doc}')

    def test_get_tokens(self):
        self._print(f"CdocsTests.test_get_tokens")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee"
        cdocs = Cdocs(PATH)
        tokens = cdocs.get_tokens(docpath)
        self._print(f"test_get_tokens: tokens: {tokens}")
        self.assertIn("company", tokens, msg=f"tokens at {docpath} must include key 'company'")
        self.assertIn("app", tokens, msg=f"tokens at {docpath} must include key 'app'")
        #self.assertIn("other", tokens, msg=f"tokens at {docpath} must include key 'other'")
        company = tokens["company"]
        self.assertEqual(company, "starstruck")

    def test_get_labels(self):
        self._print(f"CdocsTests.test_get_labels")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee"
        cdocs = Cdocs(PATH)
        labels = cdocs.get_labels(docpath)
        self._print(f"test_get_labels: the labels are: {labels}")
        self.assertIn("app", labels, msg=f"labels at {docpath} must include key 'app'")
        self.assertIn("team", labels, msg=f"labels at {docpath} must include key 'team'")
        self.assertIn("my_app_name", labels, msg=f"labels at {docpath} must include key 'my_app_name'")
        self.assertEqual("my app's name is fruit", labels["my_app_name"], msg=f"label my app name must == my app's name is fruit")
        self.assertEqual("starstruck", labels["company"], msg=f"label company must == starstruck")
        #
        # get labels from the top
        #
        self.assertIn("filenames", labels, msg=f"labels at {docpath} must include key 'filenames' from /")


    def test_get_labels_no_recurse(self):
        self._print(f"CdocsTests.test_get_labels_no_recurse")
        if self.off(): return
        docpath1 = "/v1/config/names"
        docpath2 = "/v1/config"
        cdocs = Cdocs(APIUI)

        labels = cdocs.get_labels(docpath1)
        self._print(f"test_get_labels: the labels are: {labels}")
        self.assertIn("docroot", labels, msg=f"labels at {docpath1} must include key 'docroot'")
        self.assertIn("formats", labels, msg=f"labels at {docpath1} must include key 'format'")

        labels = cdocs.get_labels(docpath1, False)
        self._print(f"test_get_labels: the labels are: {labels}")
        self.assertIn("docroot", labels, msg=f"labels at {docpath1} must include key 'docroot'")
        self.assertNotIn("formats", labels, msg=f"labels at {docpath1} must not include key 'format'")

        labels = cdocs.get_labels(docpath2, False)
        self._print(f"test_get_labels: the labels are: {labels}")
        self.assertNotIn("docroot", labels, msg=f"labels at {docpath2} must not include key 'docroot'")
        self.assertIn("formats", labels, msg=f"labels at {docpath2} must include key 'format'")

        labels = cdocs.get_labels(docpath2, True)
        self._print(f"test_get_labels: the labels are: {labels}")
        self.assertNotIn("docroot", labels, msg=f"labels at {docpath2} must not include key 'docroot'")
        self.assertIn("formats", labels, msg=f"labels at {docpath2} must include key 'format'")


    def test_get_filename(self):
        self._print(f"CdocsTests.test_get_filename")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee#new_assignee"
        cdocs = Cdocs(PATH)
        filename = cdocs._pather.get_filename(docpath)
        self._print(f"test_get_filename: filename: {filename}")
        self.assertEqual(filename, "new_assignee", msg=f"filename must be 'new_assignee'")

    def test_get_filename_special_hashmark(self):
        self._print(f"CdocsTests.test_get_filename_special_hashmark")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee.new_assignee"
        cdocs = Cdocs(PATH)
        cdocs._pather._hashmark = '.'
        filename = cdocs._pather.get_filename(docpath)
        self._print(f"test_get_filename_special_hashmark: filename: {filename}")
        self.assertEqual(filename, "new_assignee", msg=f"filename must be 'new_assignee'")

    def test_get_full_file_path(self):
        self._print(f"CdocsTests.test_get_full_file_path")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee#new_assignee"
        cdocs = Cdocs(PATH)
        path = cdocs._pather.get_full_file_path(docpath)
        found = path.find("/assignee/new_assignee.xml")
        self._print(f"test_get_full_file_path: found: {found}")
        self.assertNotEqual( -1, found, msg=f"path {path} must end in '/assignee/new_assignee.xml'")

    def test_get_concat_paths(self):
        self._print(f"CdocsTests.test_get_concat_paths")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee/concat.concat"
        cdocs = Cdocs(PATH)
        paths = cdocs._get_concat_paths(docpath)
        self.assertNotEqual(None, paths, msg=f'cannot get concat paths from {docpath}')
        self._print(f'test_get_concat_paths: paths: {paths}')
        self.assertIn( "/app/home/teams/todos/assignee#new_assignee", paths, msg=f"paths: {paths} must include in '/app/home/teams/todos/assignee#new_assignee'")

    def test_get_concat_doc(self):
        self._print(f"CdocsTests.test_get_concat_doc")
        if self.off(): return
        docpath = "/app/home/teams/todos/assignee/concat.concat"
        cdocs = Cdocs(PATH)
        doc = cdocs.get_concat_doc(docpath)
        na = doc.find("new assignee")
        ea = doc.find("edit assignee")
        self.assertNotEqual(-1, na, msg=f'{docpath} must include "new assignee" in {doc}')
        self.assertNotEqual(-1, ea, msg=f'{docpath} must include "edit assignee" in {doc}')
        self._print(f'test_get_concat_doc: concat doc: {doc}')

    def test_get_compose_doc(self):
        self._print(f"CdocsTests.test_get_compose_doc")
        #if self.off(): return
        docpath = "/app/home/teams/compose.html"
        cdocs = Cdocs(PATH)
        doc = cdocs.get_compose_doc(docpath)
        na = doc.find("new assignee")
        ea = doc.find("edit assignee")
        self.assertNotEqual(-1, na, msg=f'{docpath} must include "new assignee" in {doc}')
        self.assertNotEqual(-1, ea, msg=f'{docpath} must include "edit assignee" in {doc}')
        self._print(f'test_get_compose_doc: compose doc: {doc}')

    def test_get_doc_from_root_with_multiple_ext(self):
        self._print(f"CdocsTests.test_get_doc_from_root_with_multiple_ext")
        if self.off(): return
        docpath = "/app/home"
        cdocs = Cdocs(PATH2)
        doc = cdocs.get_doc(docpath)
        self.assertIsNotNone(doc, msg=f'{docpath} must not return None')
        home = doc.find("This is home in text!")
        self.assertNotEqual(-1, home, msg=f'{docpath} must include "This is home in text!" in {doc}')
        doc = cdocs.get_doc(docpath + "/teams")
        self.assertIsNotNone(doc, msg=f'{docpath} must not return None')
        home = doc.find("<teams>")
        self.assertNotEqual(-1, home, msg=f'{docpath} must include "<teams>" in {doc}')

    def test_concat_json(self):
        self._print(f"\n\n>>>>>>>>>>>>>>>> CdocsTests.test_concat_json")
        if self.off(): return
        docpath = "/app/home+home_screen"
        cdocs = Cdocs(JSON)
        self._print(f"CdocsTests.test_concat_json: root name: {cdocs.rootname}")
        doc = cdocs.get_doc(docpath)
        self.assertIsNotNone(doc, msg=f'{docpath} must not return None')
        self._print(f"CdocsTests.test_concat_json: doc: {doc}\n\n\n")
        home = doc.find('{"how-to-use": "How to use My Home", "header": "Home Screen", "content": "The home screen is where you land after logging in."}')
        self.assertNotEqual(-1, home, msg=f'{docpath} must include the joined json, not: {doc}')

    def test_list_docs(self):
        self._print(f"CdocsTests.test_list_docs")
        docpath = "/app/home/teams"
        cdocs = Cdocs(PATH)
        self._print(f"CdocsTests.test_list_docs. cdocs: {cdocs}")
        docs = cdocs.list_docs(docpath)
        self._print(f"CdocsTests.test_list_docs: docs: {docs}")
        self.assertIsNotNone(docs, msg=f"docpath: {docpath} must not be None")
        self.assertEqual(len(docs), 3, msg=f"len(docs) at {docpath} must be 3")
        self.assertIn("a.json", docs, msg=f"docs {docs} must include a.json")

        self._print(f"CdocsTests.test_list_docs: listing root")
        docs = cdocs.list_docs("/")
        self._print(f"CdocsTests.test_list_docs: docs: {docs}")
        self.assertIsNotNone(docs, msg=f"docpath: {docpath} must not be None")
        self.assertEqual(len(docs), 3, msg=f"len(docs) at {docpath} must be 3")
        self.assertIn("404a.xml", docs, msg=f"docs {docs} must include 404a.xml")

        self._print(f"CdocsTests.test_list_docs: listing empty string -- which is also the root")
        docs = cdocs.list_docs("")
        self._print(f"CdocsTests.test_list_docs: docs: {docs}")
        self.assertIsNotNone(docs, msg=f"docpath: {docpath} must not be None")
        self.assertEqual(len(docs), 3, msg=f"len(docs) at {docpath} must be 3")
        self.assertIn("404a.xml", docs, msg=f"docs {docs} must include 404a.xml")

        self._print(f"CdocsTests.test_list_docs: listing bogus docpath")
        docs = cdocs.list_docs("fish/can or cannot/fly")
        self._print(f"CdocsTests.test_list_docs: docs: {docs}")
        self.assertIsNotNone(docs, msg=f"docpath: {docpath} must not be None")
        self.assertEqual(len(docs), 0, msg=f"len(docs) at {docpath} must be 0")

    def test_doc_at_root(self):
        self._print(f"CdocsTests.test_doc_at_root")
        #
        # we have:
        #   /404x.html
        #   /404
        #   /404a.xml
        #
        docpath = "/#404x.html"
        cdocs = Cdocs(PATH)
        #self._debug()
        self._print(f"CdocsTests.test_doc_at_root. cdocs: {cdocs}")
        doc = cdocs.get_doc(docpath)
        self._print(f"CdocsTests.test_doc_at_root: doc: {doc}")
        self.assertIsNotNone(doc, msg=f"docpath: {docpath} must not be None")
        found = doc.find("REALLY") >= 0
        # should not be found because this root doesn't have html as a format
        # simple pather look using its exts resulting in 404x.html.xml and 404x.html.json
        self.assertEqual(found, False, msg=f"doc at {docpath} must not include 'REALLY'")









from pathlib import Path
import os
import json
import logging
from cdocs.contextual_docs import DocPath, FilePath, JsonDict
from cdocs.finder import Finder
from typing import Optional

class FinderException(Exception):
    pass

class SimpleFinder(Finder):

    def __init__(self, cdocs): #can't type hint cdocs
        self._cdocs = cdocs
        self._docroot = cdocs.get_doc_root()

    def find_tokens(self, path:DocPath=None, filename:str="tokens.json", recurse:Optional[bool]=True) -> JsonDict:
        """ first checks "public", then other roots, last "internal". prefered in that order """
        tokens = JsonDict(dict())
        if path is None:
            return tokens
        pointer = os.path.join(self._docroot, path)
        tokens = self._find_tokens( self._docroot, pointer, tokens, filename, recurse)
        return tokens

    def _find_tokens( self, root:FilePath, pointer:FilePath, tokens:JsonDict, filename:str, recurse:Optional[bool]=True ) -> JsonDict:
        if pointer == "" or pointer is None:
            raise FinderException(f"_find_tokens got bad pointer: {pointer} in {root}")
        if pointer == root:
            return tokens
        tfile = self._join(pointer, filename)
        tdict = self._read_json(tfile)
        tokens = {**tdict, **tokens}
        if recurse:
            logging.info(f"simple_finder._find_tokens: recursing on pointer: {pointer}")
            end = int( pointer.rfind("/") )
            logging.info(f"simple_finder._find_tokens: end: {end}")
            pointer = pointer[0:end]
            if pointer == root:
                pointer = pointer + "/"
                recurse = False
            logging.info(f"simple_finder._find_tokens: now pointer: {pointer}")
            return self._find_tokens(root, pointer, tokens, filename, recurse)
        else:
            return tokens

    def _join(self, pointer:FilePath, filename:str) -> FilePath:
        dot = pointer.find(".")
        if dot == -1:
            return os.path.join(pointer, filename)
        else:
            ext = pointer[pointer.rindex('.')+1:]
            p = pointer[0:pointer.rindex('/')]
            logging.info(f"SimpleFinder._join: p: {p}")
            j = os.path.join(p, filename)
            logging.info(f"SimpleFinder._join: j: {j}")
            return j

    def _read_json(self, path) -> JsonDict:
        try:
            with open(path) as f:
                return JsonDict(json.load(f))
        except FileNotFoundError as e:
            logging.debug(f"DictFinder._read_json: no such file {path}. returning empty dict, as expected.")
            return JsonDict(dict())




import abc
from typing import NewType, Dict, Callable, Union, Optional, List


# Doc is the contextual documentation content
Doc = NewType('Doc', str)

# DocPath is a relative path structure that points to
# a contextual documentation doc. A DocPath is like
# /x/y/z. It can have a leading '/' char, or not. DocPath
# are passed in from outside the library.
DocPath = NewType('DocPath', str)

# FilePath is the actual physical file path that contains Doc
# content.
FilePath = NewType('FilePath', str)

# JsonDict is a dict with string keys with either a
# string value or a function that takes a string and returns
# a string. At this time the only such function is
# ContextualDocs.get_doc. JsonDict are for tokens and labels.
# Only tokens JsonDict have callable members.
JsonDict = NewType('JsonDict', Dict[str, Union[str,Callable[[str],str]]])

# a Path is either a DocPath or a FilePath. in some cases
# a helper may need to recieve or treat as one or the other
# where another helper might need or treat the other way.
Path = NewType('Path', Union[DocPath,FilePath])


class ContextualDocs(metaclass=abc.ABCMeta):
    """
    See the README.md file for description of the library and
    these methods.
    """

    @abc.abstractmethod
    def get_concat_doc(self, path:DocPath) -> Optional[Doc]:
        pass

    @abc.abstractmethod
    def get_compose_doc(self, path:DocPath) -> Optional[Doc]:
        pass

    @abc.abstractmethod
    def get_doc(self, path:DocPath) -> Optional[Doc]:
        pass

    @abc.abstractmethod
    def get_labels(self, path:DocPath, recurse:Optional[bool]=True, addlabels:Optional[bool]=True) -> Optional[JsonDict]:
        pass

    @abc.abstractmethod
    def get_tokens(self, path:DocPath, recurse:Optional[bool]=True) -> JsonDict:
        pass

    @abc.abstractmethod
    def list_docs(self, path:DocPath) -> List[Doc]:
        pass


